/*
 * template - Homogeneous vector template
 *
 *  (c) 1999 Shiro Kawai  shiro@acm.org
 *  Terms and conditions to use, modify or distribute this code
 *  is the same as of STk.
 *
 *  $Id: uv_templ.c,v 1.4 1999/11/08 11:22:24 shiro Exp $
 */

/* Poor man's template
 *
 * The caller must define those macros:
 *
 *  TAG           `s8', `u32', etc.. 
 *  TAGTYPE       C type corresponding to the tag
 *  XVECTOR_TYPE  UvectorType
 *  XVECTORP(obj) Predicate
 *  XVECTOR_DATA(obj) macro to extract data pointer to TAGTYPE.
 *  ASSERT_TAGTYPE(obj)
 *                A macro which checks SCM obj can be converted
 *                to C's TAGTYPE.  If not, throw an error.
 *  BOX_TAGTYPE(obj, val)
 *                A macro which converts C value VAL to SCM obj
 *  UNBOX_TAGTYPE(val, obj)
 *                A macro which converts SCM obj to C value VAL
 *  COMPARE_TAGTYPE(val1, val2)
 *                Compare two C TAGTYPE values and return true
 *                if it's the same, false otherwise.
 */

#include <stdlib.h>
#include <math.h>
#include <string.h>  /* for memcpy() */
#include <stk.h>

/* 
 * Macros
 */

/* cpp weirdness */
#define _CONCAT2(a, b) a ## b
#define _CONCAT3(a, b, c) a ## b ## c

#define CONCAT2(a, b) _CONCAT2(a, b)
#define CONCAT3(a, b, c) _CONCAT3(a, b, c)

#define _SCONCAT2(a, b)    #a b
#define _SCONCAT3(a, b, c) a #b c

#define SCONCAT2(a, b)     _SCONCAT2(a, b)
#define SCONCAT3(a, b, c)  _SCONCAT3(a, b, c)

/* generic accessor */
#define XVECTOR_TYPEDESC   CONCAT2(TAG, vector_desc)
#define XVECTOR_EXTDATA    CONCAT2(TAG, vector_extdata)

#define XVECTOR_LENGTH(obj)  UVECTOR_LENGTH(obj)
#define XVECTOR_NEW          CONCAT3(STk_, TAG, vector_new)

#define ASSERT_XVECTOR(obj)  \
    if (!XVECTORP(obj)) Serror(SCONCAT2(TAG, "vector expected"), obj)

#ifndef XVECTOR_EXTRA_INIT
#define XVECTOR_EXTRA_INIT(obj)
#endif

/* procedure names */
#define C_XVECTOR_P          CONCAT3(STk_, TAG, vectorp)
#define S_XVECTOR_P          SCONCAT2(TAG, "vector?")
#define C_MAKE_XVECTOR       CONCAT3(STk_make_, TAG, vector)
#define S_MAKE_XVECTOR       SCONCAT3("make-", TAG, "vector")
#define C_XVECTOR            CONCAT3(STk_, TAG, vector)
#define S_XVECTOR            SCONCAT2(TAG, "vector")
#define C_XVECTOR_LENGTH     CONCAT3(STk_, TAG, vector_length)
#define S_XVECTOR_LENGTH     SCONCAT2(TAG, "vector-length")
#define C_XVECTOR_REF        CONCAT3(STk_, TAG, vector_ref)
#define S_XVECTOR_REF        SCONCAT2(TAG, "vector-ref")
#define C_XVECTOR_SET        CONCAT3(STk_, TAG, vector_set)
#define S_XVECTOR_SET        SCONCAT2(TAG, "vector-set!")
#define C_XVECTOR2LIST       CONCAT3(STk_, TAG, vector2list)
#define S_XVECTOR2LIST       SCONCAT2(TAG, "vector->list")
#define C_LIST2XVECTOR       CONCAT3(STk_list2, TAG, vector)
#define S_LIST2XVECTOR       SCONCAT3("list->", TAG, "vector")

#define C_XVECTOR_COPY       CONCAT3(STk_, TAG, vector_copy)
#define S_XVECTOR_COPY       SCONCAT2(TAG, "vector-copy")
#define C_XVECTOR_COPY_DESTRUCTIVE CONCAT3(STk_, TAG, vector_copy_destructive)
#define S_XVECTOR_COPY_DESTRUCTIVE SCONCAT2(TAG, "vector-copy!")
#define C_XVECTOR2VECTOR     CONCAT3(STk_, TAG, vector2vector)
#define S_XVECTOR2VECTOR     SCONCAT2(TAG, "vector->vector")
#define C_VECTOR2XVECTOR     CONCAT3(STk_vector2, TAG, vector)
#define S_VECTOR2XVECTOR     SCONCAT3("vector->", TAG, "vector")

#ifdef VECTOR_MATH
#define C_XVECTOR_NORM       CONCAT3(STk_, TAG, vector_norm)
#define S_XVECTOR_NORM       SCONCAT2(TAG, "vector-norm")
#define C_XVECTOR_NORMALIZE  CONCAT3(STk_, TAG, vector_normalize)
#define S_XVECTOR_NORMALIZE  SCONCAT2(TAG, "vector-normalize!")
#define C_XVECTOR_SCALE      CONCAT3(STk_, TAG, vector_scale)
#define S_XVECTOR_SCALE      SCONCAT2(TAG, "vector-scale")
#define C_XVECTOR_SCALE_DESTRUCTIVE CONCAT3(STk_, TAG, vector_scale_destructive)
#define S_XVECTOR_SCALE_DESTRUCTIVE SCONCAT2(TAG, "vector-scale!")
#define C_XVECTOR_ADD        CONCAT3(STk_, TAG, vector_add)
#define S_XVECTOR_ADD        SCONCAT2(TAG, "vector-add")
#define C_XVECTOR_INCREMENT  CONCAT3(STk_, TAG, vector_increment)
#define S_XVECTOR_INCREMENT  SCONCAT2(TAG, "vector-increment!")
#define C_XVECTOR_SUBTRACT   CONCAT3(STk_, TAG, vector_subtract)
#define S_XVECTOR_SUBTRACT   SCONCAT2(TAG, "vector-subtract")
#define C_XVECTOR_DECREMENT  CONCAT3(STk_, TAG, vector_decrement)
#define S_XVECTOR_DECREMENT  SCONCAT2(TAG, "vector-decrement!")
#define C_XVECTOR_DOT_PRODUCT CONCAT3(STk, TAG, vector_dot_product)
#define S_XVECTOR_DOT_PRODUCT SCONCAT2(TAG, "vector-dot-product")
#define C_XVECTOR_CROSS_PRODUCT CONCAT3(STk, TAG, vector_cross_product)
#define S_XVECTOR_CROSS_PRODUCT SCONCAT2(TAG, "vector-cross-product")

#ifndef EPSILON
#define EPSILON  10e-5
#endif /*EPSILON*/

#endif /*VECTOR_MATH*/

/*
 * Constructor
 */

SCM XVECTOR_NEW(int length)
{
    XVECTOR_EXTDATA *data =
        must_malloc(sizeof(XVECTOR_EXTDATA) + (length - 1)*sizeof(TAGTYPE));
    SCM cell;
    data->header.length = length;
    data->header.type = XVECTOR_TYPE;
    NEWCELL(cell, tc_uvector);
    EXTDATA(cell) = (void *)data;
    return cell;
}

static SCM xvector_from_list(SCM args, char *proc_name)
{
    int len = STk_llength(args);
    SCM p, v;
    TAGTYPE *data;

    if (len < 0) Serror("list expected", args);
    v = XVECTOR_NEW(len);
    data = XVECTOR_DATA(v);
    for (p=args; CONSP(p); p = CDR(p), data++) {
        SCM elt = CAR(p);
        UNBOX_TAGTYPE(*data, elt);
    }
    return v;
}

/*====================================================================
 * SRFI-4 procedures
 */

/*
 * (TAGvector? OBJ)
 */

PRIMITIVE C_XVECTOR_P(SCM obj)
{
    return (XVECTORP(obj)? Truth : Ntruth);
}

/*
 * (make-TAGvector length [init])
 */

PRIMITIVE C_MAKE_XVECTOR(SCM slen, SCM sinit)
{
    static char *proc_name = S_MAKE_XVECTOR;
    SCM v;
    TAGTYPE *data;
    int len, i;
    
    if (!EXACTP(slen))  Serror("exact number required", slen);
    if((len = STk_integer_value(slen)) < 0) {
        Serror("length out of range", slen);
    }
    if (UNBOUND != sinit) {
        ASSERT_TAGTYPE(sinit);  /* check here to detect errors */
    }
    v = XVECTOR_NEW(len);
    if (UNBOUND != sinit) {
        TAGTYPE init;
        UNBOX_TAGTYPE(init, sinit);
        data = XVECTOR_DATA(v);
        for (i = 0; i < len; i++) *data++ = init;
    } else {
        XVECTOR_EXTRA_INIT(v);
    }
    return v;
}

/*
 * (TAGvector value ...)
 */

PRIMITIVE C_XVECTOR(SCM args, int len)
{
    return xvector_from_list(args, S_XVECTOR);
}

/*
 * (TAGvector-length obj)
 */

PRIMITIVE C_XVECTOR_LENGTH(SCM obj)
{
    static char *proc_name = S_XVECTOR_LENGTH;
    int len;
    ASSERT_XVECTOR(obj);
    len = XVECTOR_LENGTH(obj);
    return STk_makeinteger(len);
}

/*
 * (TAGvector-ref obj i)
 */

PRIMITIVE C_XVECTOR_REF(SCM obj, SCM si)
{
    static char *proc_name = S_XVECTOR_REF;
    int i;
    TAGTYPE elt;
    SCM result;

    ASSERT_XVECTOR(obj);
    if (!EXACTP(si)) Serror("exact number expected", si);
    i = STk_integer_value(si);
    if (i < 0 || i >= XVECTOR_LENGTH(obj)) {
        Serror("index out of range", si);
    }
    elt = XVECTOR_DATA(obj)[i];
    BOX_TAGTYPE(result, elt);
    return result;
}

/*
 * (TAGvector-set! obj i val)
 */

PRIMITIVE C_XVECTOR_SET(SCM obj, SCM si, SCM sval)
{
    static char *proc_name = S_XVECTOR_SET;
    int i;
    TAGTYPE val;

    ASSERT_XVECTOR(obj);
    if (!EXACTP(si)) Serror("exact number expected", si);
    i = STk_integer_value(si);
    if (i < 0 || i >= XVECTOR_LENGTH(obj)) {
        Serror("index out of range", si);
    }
    UNBOX_TAGTYPE(val, sval);
    XVECTOR_DATA(obj)[i] = val;
    return UNDEFINED;
}

/*
 * (TAGvector->list obj)
 */

PRIMITIVE C_XVECTOR2LIST(SCM obj)
{
    static char *proc_name = S_XVECTOR2LIST;
    SCM head = NIL, tail = NIL;
    TAGTYPE *data;
    int i, len;

    ASSERT_XVECTOR(obj);
    len = XVECTOR_LENGTH(obj);
    data = XVECTOR_DATA(obj);
    for (i=0; i<len; i++, data++) {
        SCM elt;
        BOX_TAGTYPE(elt, *data);
        if (head == NIL) {
            tail = head = Cons(elt, NIL);
        } else {
            CDR(tail) = Cons(elt, NIL);
            tail = CDR(tail);
        }
    }
    return head;
}

/*
 * (list->TAGvector obj)
 */
PRIMITIVE C_LIST2XVECTOR(SCM obj)
{
    return xvector_from_list(obj, S_LIST2XVECTOR);
}


/*====================================================================
 * non-SRFI procedures
 */

/*
 * (TAGvector-copy obj)
 */

PRIMITIVE C_XVECTOR_COPY(SCM obj)
{
    static char *proc_name = S_XVECTOR_COPY;
    int len;
    SCM v;
    ASSERT_XVECTOR(obj);
    len = XVECTOR_LENGTH(obj);
    v = XVECTOR_NEW(len);
    memcpy(XVECTOR_DATA(v), XVECTOR_DATA(obj), len * sizeof(TAGTYPE));
    return v;
}

/*
 * (TAGvector-copy! DEST SRC)
 *   SRC and DEST must be TAGvectors with the same length.
 *   Copy contents of SRC into DEST.  Returns DEST.
 */

PRIMITIVE C_XVECTOR_COPY_DESTRUCTIVE(SCM dest, SCM src)
{
    static char *proc_name = S_XVECTOR_COPY_DESTRUCTIVE;
    int len;
    ASSERT_XVECTOR(src);
    ASSERT_XVECTOR(dest);
    if ((len = XVECTOR_LENGTH(src)) != XVECTOR_LENGTH(dest)) {
        Serror("same length of vectors are required", LIST2(src, dest));
    }
    memcpy(XVECTOR_DATA(dest), XVECTOR_DATA(src), len * sizeof(TAGTYPE));
    return dest;
}

/*
 * (TAGvector->vector VEC)
 */

PRIMITIVE C_XVECTOR2VECTOR (SCM obj)
{
    static char *proc_name = S_XVECTOR2VECTOR;
    int len, i;
    TAGTYPE *data;
    SCM sv, *sdata;
    
    ASSERT_XVECTOR(obj);
    len = XVECTOR_LENGTH(obj);
    data = XVECTOR_DATA(obj);
    sv = STk_makevect(len, NIL);
    sdata = VECT(sv);
    for (i=0; i<len; i++, data++, sdata++) {
        BOX_TAGTYPE(*sdata, *data);
    }
    return sv;
}

/*
 * (vector->TAGvector VEC)
 */

PRIMITIVE C_VECTOR2XVECTOR (SCM obj)
{
    static char *proc_name = S_VECTOR2XVECTOR;
    int len, i;
    TAGTYPE *data;
    SCM sv, *sdata;

    if (!VECTORP(obj)) Serror("vector expected", obj);
    len = VECTSIZE(obj);
    sdata = VECT(obj);
    sv = XVECTOR_NEW(len);
    data = XVECTOR_DATA(sv);
    for (i=0; i<len; i++, data++, sdata++) {
        UNBOX_TAGTYPE(*data, *sdata);
    }
    return sv;
}

#ifdef VECTOR_MATH

/*====================================================================
 * a little bit of mathematics...
 */

/*
 * (TAGvector-norm VEC)
 *   Return norm of fvector FV.
 */

PRIMITIVE C_XVECTOR_NORM (SCM obj)
{
    static char *proc_name = S_XVECTOR_NORM;
    int len, i;
    double norm = 0.0;
    TAGTYPE *data;

    ASSERT_XVECTOR(obj);
    len = XVECTOR_LENGTH(obj);
    data = XVECTOR_DATA(obj);
    for (i=0; i<len; i++) {
        double v = (double)*data++;
        norm += v*v;
    }
    norm = sqrt(norm);
    return STk_makenumber(norm);
}

/*
 * (TAGvector-normalize! VEC)
 *   Normalize VEC.  VEC is modified.
 *   If VEC is 
 */

PRIMITIVE C_XVECTOR_NORMALIZE (SCM obj)
{
    static char *proc_name = S_XVECTOR_NORMALIZE;
    int len, i;
    double norm = 0.0;
    TAGTYPE *data;

    ASSERT_XVECTOR(obj);
    len = XVECTOR_LENGTH(obj);
    data = XVECTOR_DATA(obj);
    for (i=0; i<len; i++) {
        double v = (double)*data++;
        norm += v * v;
    }
    norm = sqrtf(norm);
    if (fabs(norm) < EPSILON) {
        Serror("tried to normalize zero-length vector", obj);
    } else {
        data = XVECTOR_DATA(obj);
        for (i=0; i<len; i++, data++) {
            *data /= norm;
        }
    }
    return obj;
}

/* (TAGvector-scale VEC FACTOR)
 */

PRIMITIVE C_XVECTOR_SCALE (SCM fv, SCM factor)
{
    static char *proc_name = S_XVECTOR_SCALE;
    SCM nfv;
    int len;
    TAGTYPE ff, *nfp, *fp;
    
    ASSERT_XVECTOR(fv);
    if (!FLONUMP(factor)) Serror("inexact number expected", factor);
    len = XVECTOR_LENGTH(fv);
    nfv = XVECTOR_NEW(len);
    nfp = XVECTOR_DATA(nfv);
    fp = XVECTOR_DATA(fv);
    ff = (TAGTYPE)STk_real2double(factor);
    while (len--) {
        *nfp++ = ff * *fp++;
    }
    return nfv;
}

/* (TAGvector-scale! FV FACTOR)
 */

PRIMITIVE C_XVECTOR_SCALE_DESTRUCTIVE (SCM fv, SCM factor)
{
    static char *proc_name = S_XVECTOR_SCALE_DESTRUCTIVE;
    int len;
    TAGTYPE ff, *fp;
    
    ASSERT_XVECTOR(fv);
    if (!FLONUMP(factor)) Serror("inexact number expected", factor);
    len = XVECTOR_LENGTH(fv);
    fp = XVECTOR_DATA(fv);
    ff = (TAGTYPE)STk_real2double(factor);
    while (len--) {
        *fp++ *= ff;
    }
    return fv;
}

/* (TAGvector-add FV ...)
 */

PRIMITIVE C_XVECTOR_ADD (SCM args, int nargs)
{
    static char *proc_name = S_XVECTOR_ADD;
    int i, len;
    TAGTYPE *p1, *pr;
    SCM p, fr;

    if (nargs < 1) Serror("at least 1 argument expected", args);
    ASSERT_XVECTOR(CAR(args));
    len = XVECTOR_LENGTH(CAR(args));
    for (p=CDR(args); CONSP(p); p = CDR(p)) {
        ASSERT_XVECTOR(CAR(p));
        if (XVECTOR_LENGTH(CAR(p)) != len) {
            Serror(SCONCAT3("same length of ", TAG, "vectors are required"),
                   CAR(p));
        }
    }
    fr = XVECTOR_NEW(len);
    pr = XVECTOR_DATA(fr);
    p1 = XVECTOR_DATA(CAR(args));
    for (i=0; i<len; i++) *pr++ = *p1++;
    for (p=CDR(args); CONSP(p); p=CDR(p)) {
        pr = XVECTOR_DATA(fr);
        p1 = XVECTOR_DATA(CAR(p));
        for (i=0; i<len; i++) *pr++ += *p1++;
    }
    return fr;
}

/* (TAGvector-incement! FV1 FV2 ...)
 */

PRIMITIVE C_XVECTOR_INCREMENT (SCM args, int nargs)
{
    static char *proc_name = S_XVECTOR_INCREMENT;
    int i, len;
    TAGTYPE *p1, *pr;
    SCM p;

    if (nargs < 1) Serror("at least 1 argument expected", args);
    ASSERT_XVECTOR(CAR(args));
    len = XVECTOR_LENGTH(CAR(args));
    for (p=CDR(args); CONSP(p); p = CDR(p)) {
        ASSERT_XVECTOR(CAR(p));
        if (XVECTOR_LENGTH(CAR(p)) != len) {
            Serror(SCONCAT3("same length of ", TAG, "vectors are required"),
                            CAR(p));
        }
    }
    for (p=CDR(args); CONSP(p); p=CDR(p)) {
        pr = XVECTOR_DATA(CAR(args));
        p1 = XVECTOR_DATA(CAR(p));
        for (i=0; i<len; i++) *pr++ += *p1++;
    }
    return CAR(args);
}

/* (TAGvector-subtract FV1 FV2 ...)
 */

PRIMITIVE C_XVECTOR_SUBTRACT (SCM args, int nargs)
{
    static char *proc_name = S_XVECTOR_SUBTRACT;
    int i, len;
    TAGTYPE *p1, *pr;
    SCM p, fr;

    if (nargs < 1) Serror("at least 1 argument expected", args);
    ASSERT_XVECTOR(CAR(args));
    len = XVECTOR_LENGTH(CAR(args));
    for (p=CDR(args); CONSP(p); p = CDR(p)) {
        ASSERT_XVECTOR(CAR(p));
        if (XVECTOR_LENGTH(CAR(p)) != len) {
            Serror(SCONCAT3("same length of ", TAG, "vectors are required"),
                            CAR(p));
        }
    }
    fr = XVECTOR_NEW(len);
    pr = XVECTOR_DATA(fr);
    p1 = XVECTOR_DATA(CAR(args));
    for (i=0; i<len; i++) *pr++ = *p1++;
    for (p=CDR(args); CONSP(p); p=CDR(p)) {
        pr = XVECTOR_DATA(fr);
        p1 = XVECTOR_DATA(CAR(p));
        for (i=0; i<len; i++) *pr++ -= *p1++;
    }
    return fr;
}

/* (TAGvector-decrement! FV1 FV2 ...)
 */

PRIMITIVE C_XVECTOR_DECREMENT (SCM args, int nargs)
{
    static char *proc_name = S_XVECTOR_DECREMENT;
    int i, len;
    TAGTYPE *p1, *pr;
    SCM p;

    if (nargs < 1) Serror("at least 1 argument expected", args);
    ASSERT_XVECTOR(CAR(args));
    len = XVECTOR_LENGTH(CAR(args));
    for (p=CDR(args); CONSP(p); p = CDR(p)) {
        ASSERT_XVECTOR(CAR(p));
        if (XVECTOR_LENGTH(CAR(p)) != len) {
            Serror(SCONCAT3("same length of ", TAG, "vectors are required"),
                           CAR(p));
        }
    }
    for (p=CDR(args); CONSP(p); p=CDR(p)) {
        pr = XVECTOR_DATA(CAR(args));
        p1 = XVECTOR_DATA(CAR(p));
        for (i=0; i<len; i++) *pr++ -= *p1++;
    }
    return CAR(args);
}

/* (TAGvector-dot-product FV1 FV2)
 */

PRIMITIVE C_XVECTOR_DOT_PRODUCT (SCM fv1, SCM fv2)
{
    static char *proc_name = S_XVECTOR_DOT_PRODUCT;
    int i, len;
    TAGTYPE value = (TAGTYPE)0.0;
    TAGTYPE *p1, *p2;
    ASSERT_XVECTOR(fv1);
    ASSERT_XVECTOR(fv2);
    if ((len = XVECTOR_LENGTH(fv1)) != XVECTOR_LENGTH(fv2)) {
        Serror(SCONCAT3("same length of ", TAG, "vectors are required"),
                        Cons(fv1, fv2));
    }
    p1 = XVECTOR_DATA(fv1);
    p2 = XVECTOR_DATA(fv2);
    for (i=0; i<len; i++) {
        value += *p1++ * *p2++;
    }
    return STk_makenumber((double)value);
}

/* (TAGvector-cross-product FV1 FV2)
 *   Calculate cross product of FV1 and FV2.  Length of both
 *   arguments must be 3.
 */

PRIMITIVE C_XVECTOR_CROSS_PRODUCT (SCM fv1, SCM fv2)
{
    static char *proc_name = S_XVECTOR_CROSS_PRODUCT;
    SCM fvr;
    TAGTYPE *pr, *p1, *p2;
    ASSERT_XVECTOR(fv1);
    ASSERT_XVECTOR(fv2);
    if (XVECTOR_LENGTH(fv1) != 3) {
        Serror("vector length must be 3", fv1);
    }
    if (XVECTOR_LENGTH(fv2) != 3) {
        Serror("vector length must be 3", fv2);
    }
    fvr = XVECTOR_NEW(3);
    pr = XVECTOR_DATA(fvr);
    p1 = XVECTOR_DATA(fv1);
    p2 = XVECTOR_DATA(fv2);
    pr[0] = p1[1]*p2[2] - p1[2]*p2[1];
    pr[1] = p1[2]*p2[0] - p1[0]*p2[2];
    pr[2] = p1[0]*p2[1] - p1[1]*p2[0];
    return fvr;
}

#endif /*VECTOR_MATH*/

/*===========================================================
 * Initialization
 */

/*
 * Extended type data structure
 */

static void CONCAT2(TAG, vector_display)(SCM obj, SCM port, int mode)
{
    char buf[50];
    TAGTYPE *data = XVECTOR_DATA(obj);
    int i, len = XVECTOR_LENGTH(obj);
    Puts(SCONCAT3("#", TAG, "("), port);
    for (i=0; i < len; i++) {
        TAGTYPE val = *data++;
        if (i>0) Putc(' ', port);
        PUTS_TAGTYPE(val, buf, port);
    }
    Puts(")", port);
}

static SCM CONCAT2(TAG, vector_compare)(SCM a, SCM b, int equalp)
{
    int i;
    if (!XVECTORP(a) || !XVECTORP(b)) return Ntruth;
    if (XVECTOR_LENGTH(a) != XVECTOR_LENGTH(b)) return Ntruth;
    for (i=0; i<XVECTOR_LENGTH(a); i++) {
        TAGTYPE ca = XVECTOR_DATA(a)[i];
        TAGTYPE cb = XVECTOR_DATA(b)[i];
        if (!COMPARE_TAGTYPE(ca, cb)) return Ntruth;
    }
    return Truth;
}

uvector_subtype XVECTOR_TYPEDESC = {
    SCONCAT2(TAG, ""),            /* name */
    XVECTOR_MARK,                 /* mark */
    CONCAT2(TAG, vector_display), /* display */
    CONCAT2(TAG, vector_compare), /* compare */
    C_MAKE_XVECTOR,               /* make_from_length */
    C_XVECTOR                     /* make_from_args */
};


void CONCAT2(TAG, vector_init)() 
{
    STk_add_new_primitive(S_XVECTOR_P, tc_subr_1,
                          (PRIMITIVE(*)())C_XVECTOR_P);
    STk_add_new_primitive(S_MAKE_XVECTOR, tc_subr_1_or_2,
                          (PRIMITIVE(*)())C_MAKE_XVECTOR);
    STk_add_new_primitive(S_XVECTOR, tc_lsubr,
                          (PRIMITIVE(*)())C_XVECTOR);
    STk_add_new_primitive(S_XVECTOR_LENGTH, tc_subr_1,
                          (PRIMITIVE(*)())C_XVECTOR_LENGTH);
    STk_add_new_primitive(S_XVECTOR_REF, tc_subr_2,
                          (PRIMITIVE(*)())C_XVECTOR_REF);
    STk_add_new_primitive(S_XVECTOR_SET, tc_subr_3,
                          (PRIMITIVE(*)())C_XVECTOR_SET);
    STk_add_new_primitive(S_XVECTOR2LIST, tc_subr_1,
                          (PRIMITIVE(*)())C_XVECTOR2LIST);
    STk_add_new_primitive(S_LIST2XVECTOR, tc_subr_1,
                          (PRIMITIVE(*)())C_LIST2XVECTOR);

    STk_add_new_primitive(S_XVECTOR_COPY, tc_subr_1,
                          (PRIMITIVE(*)())C_XVECTOR_COPY);
    STk_add_new_primitive(S_XVECTOR_COPY_DESTRUCTIVE, tc_subr_2,
                          (PRIMITIVE(*)())C_XVECTOR_COPY_DESTRUCTIVE);
    STk_add_new_primitive(S_XVECTOR2VECTOR, tc_subr_1,
                          (PRIMITIVE(*)())C_XVECTOR2VECTOR);
    STk_add_new_primitive(S_VECTOR2XVECTOR, tc_subr_1,
                          (PRIMITIVE(*)())C_VECTOR2XVECTOR);

#ifdef VECTOR_MATH
    STk_add_new_primitive(S_XVECTOR_NORM, tc_subr_1,
                          (PRIMITIVE(*)())C_XVECTOR_NORM);
    STk_add_new_primitive(S_XVECTOR_NORMALIZE, tc_subr_1,
                          (PRIMITIVE(*)())C_XVECTOR_NORMALIZE);
    STk_add_new_primitive(S_XVECTOR_SCALE, tc_subr_2,
                          (PRIMITIVE(*)())C_XVECTOR_SCALE);
    STk_add_new_primitive(S_XVECTOR_SCALE_DESTRUCTIVE, tc_subr_2,
                          (PRIMITIVE(*)())C_XVECTOR_SCALE_DESTRUCTIVE);
    STk_add_new_primitive(S_XVECTOR_ADD, tc_lsubr,
                          (PRIMITIVE(*)())C_XVECTOR_ADD);
    STk_add_new_primitive(S_XVECTOR_INCREMENT, tc_lsubr,
                          (PRIMITIVE(*)())C_XVECTOR_INCREMENT);
    STk_add_new_primitive(S_XVECTOR_SUBTRACT, tc_lsubr,
                          (PRIMITIVE(*)())C_XVECTOR_SUBTRACT);
    STk_add_new_primitive(S_XVECTOR_DECREMENT, tc_lsubr,
                          (PRIMITIVE(*)())C_XVECTOR_DECREMENT);
    STk_add_new_primitive(S_XVECTOR_DOT_PRODUCT, tc_subr_2,
                          (PRIMITIVE(*)())C_XVECTOR_DOT_PRODUCT);
    STk_add_new_primitive(S_XVECTOR_CROSS_PRODUCT, tc_subr_2,
                          (PRIMITIVE(*)())C_XVECTOR_CROSS_PRODUCT);
#endif /* VECTOR_MATH */
}
